from datetime import datetime, timedelta
import math
from typing import Any, Union

import numpy
from pyproj import CRS, Geod, Transformer

from .parsing import parse_raw_aprs

DEFAULT_CRS = CRS.from_epsg(4326)


class LocationPacket:
    """ location packet encoding (x, y, z) and time """

    def __init__(self, time: datetime, x: float, y: float, z: float = None, crs: CRS = None, source: str = None,
                 **kwargs):
        self.time = time
        self.coordinates = numpy.array((x, y, z if z is not None else 0))
        self.crs = crs if crs is not None else DEFAULT_CRS
        self.source = source
        self.attributes = kwargs

    def distance(self, point: (float, float)) -> float:
        """
        horizontal distance over ellipsoid

        :param point: (x, y) point
        :return: distance in ellipsodal units
        """

        if not isinstance(point, numpy.ndarray):
            point = numpy.array(point)
        coordinates = numpy.stack([self.coordinates[:2], point], axis=0)
        if self.crs.is_projected:
            return numpy.hypot(*numpy.sum(numpy.diff(coordinates, axis=0), axis=0))
        else:
            ellipsoid = self.crs.datum.to_json_dict()['ellipsoid']
            geodetic = Geod(a=ellipsoid['semi_major_axis'], rf=ellipsoid['inverse_flattening'])
            return geodetic.line_length(coordinates[:, 0], coordinates[:, 1])

    def transform_to(self, crs: CRS):
        transformer = Transformer.from_crs(self.crs, crs)
        self.coordinates = transformer.transform(self.coordinates)

    def __getitem__(self, field: str) -> Any:
        if field not in self:
            raise KeyError(f'"{field}" not in packet')
        return {**self.__dict__, **self.attributes}[field]

    def __setitem__(self, field: str, value: Any):
        self.attributes[field] = value

    def __contains__(self, field: str) -> bool:
        return field in {**self.__dict__, **self.attributes}

    def __sub__(self, other: 'LocationPacket') -> 'Distance':
        return Distance.from_packets(self, other)

    def __eq__(self, other: 'LocationPacket') -> bool:
        return numpy.allclose(self.coordinates, other.coordinates)

    def __gt__(self, other: 'LocationPacket') -> bool:
        return self.time > other.time

    def __lt__(self, other: 'LocationPacket') -> bool:
        return self.time < other.time

    def __str__(self) -> str:
        return f'{self.time} {self.coordinates}'


class Distance:
    def __init__(self, interval: timedelta, horizontal_distance: float, vertical_distance: float, crs: CRS):
        self.__interval = interval
        self.__horizontal_distance = horizontal_distance
        self.__vertical_distance = vertical_distance
        self.__crs = crs

    @classmethod
    def from_packets(cls, packet_1: LocationPacket, packet_2: LocationPacket):
        """
        Get distance between two location packets, using the first packet's CRS.

        :param packet_1: first location packet
        :param packet_2: second location packet
        """

        if packet_2.crs != packet_1.crs:
            transformer = Transformer.from_crs(packet_2.crs, packet_1.crs)
            packet_2_coordinates = transformer.transform(*packet_2.coordinates)
        else:
            packet_2_coordinates = packet_2.coordinates

        interval = packet_1.time - packet_2.time
        horizontal_distance = packet_1.distance(packet_2_coordinates[:2])
        vertical_distance = packet_1.coordinates[2] - packet_2_coordinates[2]

        return cls(interval, horizontal_distance, vertical_distance, packet_1.crs)

    @property
    def interval(self) -> timedelta:
        return self.__interval

    @property
    def seconds(self) -> float:
        return self.interval / timedelta(seconds=1)

    @property
    def horizontal_distance(self) -> float:
        return self.__horizontal_distance

    @property
    def vertical_distance(self) -> float:
        return self.__vertical_distance

    @property
    def crs(self) -> CRS:
        return self.__crs

    @property
    def distance(self) -> float:
        # TODO account for ellipsoid
        return numpy.hypot(self.__horizontal_distance, self.__vertical_distance)

    @property
    def ascent_rate(self) -> float:
        return self.__vertical_distance / self.seconds if self.seconds > 0 else math.inf

    @property
    def ground_speed(self) -> float:
        return self.__horizontal_distance / self.seconds if self.seconds > 0 else math.inf

    def __str__(self) -> str:
        return f'{self.seconds}s, {self.vertical_distance:6.2f}m vertical, {self.horizontal_distance:6.2f}m horizontal'

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({repr(self.interval)}, {self.vertical_distance}, {self.horizontal_distance}, ' \
               f'{repr(self.crs)})'


class APRSPacket(LocationPacket):
    """ APRS packet containing parsed APRS fields, along with location and time """

    def __init__(self, time: datetime, x: float, y: float, z: float, crs: CRS = None, **kwargs):
        """
        APRS packet object from raw packet and given datetime

        :param x: x value
        :param y: y value
        :param z: z value
        :param crs: coordinate reference system
        :param time: time of packet, either as datetime object, seconds since Unix epoch, or ISO format date string.
        """

        super().__init__(time, x, y, z, crs, **kwargs)

    @classmethod
    def from_raw_aprs(cls, raw_aprs: Union[str, bytes, dict], time: datetime = None, **kwargs) -> 'APRSPacket':
        """
        APRS packet object from raw packet and given datetime

        :param raw_aprs: string containing raw packet
        :param time: Time of packet, either as datetime object, seconds since Unix epoch, or ISO format date string.
        """

        # parse packet with metric units
        parsed_packet = parse_raw_aprs(raw_aprs)

        if 'longitude' in parsed_packet and 'latitude' in parsed_packet:
            if time is None:
                if 'timestamp' in parsed_packet:
                    # extract time from Unix epoch
                    time = datetime.fromtimestamp(float(parsed_packet['timestamp']))
                else:
                    # TODO make HABduino add timestamp to packet upon transmission
                    # otherwise default to time the packet was received (now)
                    time = datetime.now()

            return cls(time, parsed_packet['longitude'], parsed_packet['latitude'],
                       parsed_packet['altitude'] if 'altitude' in parsed_packet else None, crs=DEFAULT_CRS, **parsed_packet,
                       **kwargs)
        else:
            raise ValueError(f'Input packet does not contain location data: {raw_aprs}')

    @property
    def callsign(self) -> str:
        return self['callsign']

    def __getitem__(self, field: str):
        if field == 'callsign':
            field = 'from'
        return super().__getitem__(field)

    def __contains__(self, field: str):
        if field == 'callsign':
            field = 'from'
        return super().__contains__(field)

    def __iter__(self):
        yield from self.attributes

    def __eq__(self, other: 'APRSPacket') -> bool:
        return super().__eq__(other) and self.callsign == other.callsign and self['comment'] == other['comment']

    def __str__(self) -> str:
        return f'{self["callsign"]} {super().__str__()} "{self["comment"]}"'

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}(**{repr(self.coordinates)}, {self.crs.__class__.__name__}.from_wkt(' \
               f'{repr(self.crs.to_wkt())}), **{repr(self.attributes)})'



from datetime import datetime
from logging import Logger
from os import PathLike

from aprslib.packets.base import APRSPacket

from packetraven import APRSPacketDatabaseTable
from packetraven.connections import APRSPacketConnection
from packetraven.tracks import APRSTrack
from packetraven.utilities import get_logger
from packetraven.writer import write_aprs_packet_tracks

LOGGER = get_logger('packetraven')


def retrieve_packets(connections: [APRSPacketConnection], packet_tracks: [APRSTrack], database: APRSPacketDatabaseTable = None,
                     output_filename: PathLike = None, start_date: datetime = None, end_date: datetime = None,
                     logger: Logger = None) -> [APRSPacket]:
    if logger is None:
        logger = LOGGER

    logger.debug(f'receiving packets from {len(connections)} source(s)')

    parsed_packets = []
    for connection in connections:
        parsed_packets.extend(connection.packets)

    logger.debug(f'received {len(parsed_packets)} packets')

    if len(parsed_packets) > 0:
        updated_callsigns = set()
        new_packets = {}
        for parsed_packet in parsed_packets:
            callsign = parsed_packet['callsign']

            if start_date is not None and parsed_packet.time <= start_date:
                continue
            if end_date is not None and parsed_packet.time >= end_date:
                continue

            if callsign not in packet_tracks:
                packet_tracks[callsign] = APRSTrack(callsign, [parsed_packet])
                logger.debug(f'started tracking {callsign:8}')
            else:
                packet_track = packet_tracks[callsign]
                if parsed_packet not in packet_track:
                    packet_track.append(parsed_packet)
                else:
                    if database is None or parsed_packet.source != database.location:
                        logger.debug(f'skipping duplicate packet: {parsed_packet}')
                    continue

            if parsed_packet.source not in new_packets:
                new_packets[parsed_packet.source] = []
            new_packets[parsed_packet.source].append(parsed_packet)
            updated_callsigns.add(callsign)

        parsed_packets = []
        for source, packets in new_packets.items():
            logger.info(f'received {len(packets)} new packet(s) from {source}: {packets}')
            parsed_packets.extend(packets)

        parsed_packets = sorted(parsed_packets)
        updated_callsigns = sorted(updated_callsigns)

        if database is not None:
            new_packets = [packet for packet in parsed_packets if packet not in database]
            if len(new_packets) > 0:
                logger.info(f'sending {len(new_packets)} packet(s) to {database.location}')
                database.insert(new_packets)

        for callsign in updated_callsigns:
            ascent_rate = packet_tracks[callsign].ascent_rate[-1]
            ground_speed = packet_tracks[callsign].ground_speed[-1]
            seconds_to_impact = packet_tracks[callsign].seconds_to_impact
            logger.info(f'{callsign:8} ascent rate      : {ascent_rate} m/s')
            logger.info(f'{callsign:8} ground speed     : {ground_speed} m/s')
            if seconds_to_impact >= 0:
                logger.info(f'{callsign:8} estimated landing: {seconds_to_impact} s')

        if output_filename is not None:
            write_aprs_packet_tracks([packet_tracks[callsign] for callsign in updated_callsigns], output_filename)

    return parsed_packets



from ast import literal_eval
from datetime import date, datetime
from functools import partial
from socket import socket
from typing import Any, Sequence, Union

import psycopg2
from pyproj import CRS
from shapely import wkb, wkt
from shapely.errors import WKBReadingError, WKTReadingError
from shapely.geometry import LineString, LinearRing, MultiLineString, MultiPoint, MultiPolygon, Point, Polygon, \
    shape as shapely_shape
from shapely.geometry.base import BaseGeometry
from sshtunnel import SSHTunnelForwarder

from .utilities import get_logger

LOGGER = get_logger('packetraven.connection')

POSTGRES_DEFAULT_PORT = 5432
SSH_DEFAULT_PORT = 22

POSTGRES_TYPES = {
    'NoneType' : 'NULL',
    'bool'     : 'BOOL',
    'float'    : 'REAL',
    'int'      : 'INTEGER',
    'str'      : 'VARCHAR',
    'bytes'    : 'BYTEA',
    'date'     : 'DATE',
    'time'     : 'TIME',
    'datetime' : 'TIMESTAMP',
    'timedelta': 'INTERVAL',
    'list'     : 'VARCHAR[]',
    'dict'     : 'HSTORE',
    'ipaddress': 'INET',
    'Point'    : 'GEOMETRY'
}

DEFAULT_CRS = CRS.from_epsg(4326)
GEOMETRY_TYPES = (Point, LineString, Polygon, MultiPoint, MultiLineString, MultiPolygon)


class DatabaseTable:
    def __init__(self, hostname: str, database: str, table: str, fields: {str: type}, primary_key: str = None, crs: CRS = None,
                 username: str = None, password: str = None, users: [str] = None, **kwargs):
        self.hostname = hostname
        self.database = database
        self.table = table

        self.fields = fields
        if not isinstance(primary_key, Sequence) or isinstance(primary_key, str):
            primary_key = [primary_key]
        self.primary_key = primary_key
        self.crs = crs if crs is not None else DEFAULT_CRS

        if '@' in self.hostname:
            self.username, self.hostname = self.hostname.split('@')

        self.username = username
        self.password = password
        if users is None:
            users = []

        # parse port from URL
        self.hostname, self.port = split_URL_port(hostname)
        if self.port is None:
            self.port = POSTGRES_DEFAULT_PORT

        connector = partial(psycopg2.connect, database=self.database, user=self.username, password=self.password)
        if 'ssh_hostname' in kwargs:
            ssh_hostname, ssh_port = split_URL_port(kwargs['ssh_hostname'])
            if ssh_port is None:
                ssh_port = SSH_DEFAULT_PORT
            if '@' in ssh_hostname:
                ssh_username, ssh_hostname = ssh_hostname.split('@')
            ssh_username = kwargs['ssh_username'] if 'ssh_username' in kwargs else None
            ssh_password = kwargs['ssh_password'] if 'ssh_password' in kwargs else None
            self.tunnel = SSHTunnelForwarder((ssh_hostname, ssh_port),
                                             ssh_username=ssh_username, ssh_password=ssh_password,
                                             remote_bind_address=('localhost', self.port),
                                             local_bind_address=('localhost', open_tcp_port()))
            self.tunnel.start()
            self.connection = connector(host=self.tunnel.local_bind_host, port=self.tunnel.local_bind_port)
        else:
            self.tunnel = None
            self.connection = connector(host=self.hostname, port=self.port)

        with self.connection:
            with self.connection.cursor() as cursor:
                if database_has_table(cursor, self.table):
                    if database_table_is_inherited(cursor, self.table):
                        raise InheritedTableError(f'inheritance of table "{self.database}/{self.table}" '
                                                  f'will cause unexpected behaviour; aborting')

                    remote_fields = self.remote_fields
                    if list(remote_fields) != list(self.fields):
                        LOGGER.warning(f'schema of existing table "{self.database}/{self.table}" differs from given fields')

                        remote_fields_not_in_local_table = {field: value for field, value in remote_fields.items() if
                                                            field not in self.fields}
                        if len(remote_fields_not_in_local_table) > 0:
                            LOGGER.warning(f'remote table has {len(remote_fields_not_in_local_table)} fields '
                                           f'not in local table: {list(remote_fields_not_in_local_table)}')
                            LOGGER.warning(f'adding {len(remote_fields_not_in_local_table)} fields '
                                           f'to local table: {list(remote_fields_not_in_local_table)}')

                            for field, field_type in remote_fields_not_in_local_table.items():
                                previous_field = list(remote_fields)[list(remote_fields).index(field) - 1]

                                local_fields = {}
                                for local_field, local_value in self.fields.items():
                                    local_fields[local_field] = local_value
                                    if local_field == previous_field:
                                        local_fields[field] = field_type

                                self.fields = local_fields

                        local_fields_not_in_remote_table = {field: value for field, value in self.fields.items()
                                                            if field not in remote_fields}
                        if len(local_fields_not_in_remote_table) > 0:
                            LOGGER.warning(f'local table has {len(local_fields_not_in_remote_table)} fields '
                                           f'not in remote table: {list(local_fields_not_in_remote_table)}')
                            LOGGER.warning(f'adding {len(local_fields_not_in_remote_table)} fields '
                                           f'to remote table: {list(local_fields_not_in_remote_table)}')

                        if list(remote_fields) != list(self.fields):
                            LOGGER.warning(f'altering schema of "{self.database}/{self.table}"')
                            LOGGER.debug(self.remote_fields)
                            LOGGER.debug(self.fields)

                            copy_table_name = f'old_{self.table}'

                            if database_has_table(cursor, copy_table_name):
                                cursor.execute(f'DROP TABLE {copy_table_name};')

                            cursor.execute(f'ALTER TABLE {self.table} RENAME TO {copy_table_name};')

                            cursor.execute(f'CREATE TABLE {self.table} ({self.schema});')
                            for user in users:
                                cursor.execute(f'GRANT INSERT, SELECT, UPDATE, DELETE ON TABLE public.{self.table} TO {user};')

                            cursor.execute('SELECT column_name FROM information_schema.columns WHERE table_name=%s;',
                                           [f'{copy_table_name}'])
                            column_names = [record[0] for record in cursor.fetchall()]

                            cursor.execute(f'INSERT INTO {self.table} ({", ".join(column_names)}) '
                                           f'SELECT * FROM {copy_table_name};')

                            cursor.execute(f'DROP TABLE {copy_table_name};')
                else:
                    LOGGER.debug(f'creating remote table "{self.database}/{self.table}"')
                    cursor.execute(f'CREATE TABLE {self.table} ({self.schema});')

                    for user in users:
                        cursor.execute(f'GRANT INSERT, SELECT, UPDATE, DELETE ON TABLE public.{self.table} TO {user};')

    def __getitem__(self, key: Any) -> {str: Any}:
        """
        Query table for the given value of the primary key.

        Parameters
        ----------
        key
            value to query from primary key

        Returns
        -------
        {str: Any}
            dictionary record
        """

        if not isinstance(key, Sequence) or isinstance(key, str):
            key = [key]

        where = dict(zip(self.primary_key, key))
        records = self.records_where(where)

        if len(records) > 1:
            where_string = ' AND '.join(f'{key} = {value}' for key, value in where.items())
            LOGGER.warning(f'found more than one record with primary key "{where_string}": {records}')

        return records[0]

    def __setitem__(self, key: Any, record: {str: Any}):
        """
        Insert the given record into the table with the given primary key value.

        Parameters
        ----------
        key
            value of primary key at which to insert record
        record
            dictionary record
        """

        if not isinstance(key, Sequence) or isinstance(key, str):
            key = [key]

        for key_index, primary_key in enumerate(self.primary_key):
            record[primary_key] = key[key_index]

        self.insert([record])

    @property
    def records(self) -> [{str: Any}]:
        return self.records_where(None)

    def records_where(self, where: {str: Union[Any, list]}) -> [{str: Any}]:
        """
        records in the table that match the given key-value pairs

        Parameters
        ----------
        where
            dictionary mapping keys to values, with which to match records

        Returns
        -------
        [{str: Any}]
            dictionaries of matching records
        """

        matching_records = []

        with self.connection:
            with self.connection.cursor() as cursor:
                cursor.execute(f'SELECT {", ".join(self.fields.keys())} FROM {self.table}')
                records = cursor.fetchall()

        records = [parse_record_values(dict(zip(self.fields.keys(), record)), self.fields) for record in records]

        if where is None:
            matching_records = records
        else:
            try:
                for record in records:
                    match = True
                    for key, value in where.items():
                        if key in record:
                            record_value = record[key]
                            if type(value) in [tuple, list, slice, range]:
                                if record_value not in value:
                                    match = False
                            elif value != record_value if value is not None else value is not record_value:
                                match = False
                        else:
                            match = False
                    if match:
                        matching_records.append(record)
            except KeyError:
                LOGGER.warning(f'no records found with given constraints: {where}')

        return matching_records

    def insert(self, records: [{str: Any}]):
        """
        Insert the list of records into the table.

        Parameters
        ----------
        records
            dictionary records
        """

        if type(records) is dict:
            records = [records]

        assert all(primary_key in record for primary_key in self.primary_key for record in records), \
            f'one or more records does not contain primary key "{self.primary_key}"'

        with self.connection:
            with self.connection.cursor() as cursor:
                for record in records:
                    if len(self.primary_key) == 1:
                        primary_key_string = self.primary_key[0]
                        primary_key_value = record[self.primary_key[0]]
                    else:
                        primary_key_string = f'({", ".join(self.primary_key)})'
                        primary_key_value = tuple(record[primary_key] for primary_key in self.primary_key)

                    record_fields_not_in_local_table = [field for field in record if field not in self.fields]
                    if len(record_fields_not_in_local_table) > 0:
                        LOGGER.warning(f'record has {len(record_fields_not_in_local_table)} fields not in the local table '
                                       f'that will not be inserted: {record_fields_not_in_local_table}')

                    local_fields_in_record = [field for field in self.fields if field in record]
                    geometry_fields = [field for field in self.geometry_fields if field in record]

                    columns = [field for field in local_fields_in_record if field not in geometry_fields]
                    values = [record[field] for field in local_fields_in_record if field not in geometry_fields]

                    if database_table_has_record(cursor, self.table, record, self.primary_key):
                        record_without_primary_key = {column: value for column, value in zip(columns, values)
                                                      if column not in self.primary_key}
                        if len(record_without_primary_key) > 0:
                            if len(record_without_primary_key) > 1:
                                cursor.execute(f'UPDATE {self.table} SET ({", ".join(record_without_primary_key.keys())}) = %s'
                                               f' WHERE {primary_key_string} = %s;',
                                               [tuple(record_without_primary_key.values()), primary_key_value])
                            else:
                                cursor.execute(f'UPDATE {self.table} SET {tuple(record_without_primary_key.keys())[0]} = %s'
                                               f' WHERE {primary_key_string} = %s;',
                                               [tuple(record_without_primary_key.values())[0], primary_key_value])
                    else:
                        cursor.execute(f'INSERT INTO {self.table} ({", ".join(columns)}) VALUES %s;',
                                       [tuple(values)])

                    if len(geometry_fields) > 0:
                        geometries = {field: record[field] for field in geometry_fields}

                        for field, geometry in geometries.items():
                            cursor.execute(f'UPDATE {self.table} SET {field} = ST_GeomFromWKB(%s::geometry, %s) '
                                           f'WHERE {primary_key_string} = %s;',
                                           [geometry.wkb, self.crs.to_epsg(), primary_key_value])

    @property
    def schema(self) -> str:
        """ PostGres schema string of local table, given field names and types """

        schema = []
        for field, field_type in self.fields.items():
            dimensions = 0
            while type(field_type) is list:
                field_type = field_type[0]
                dimensions += 1

            field_definition = f'{field} {POSTGRES_TYPES[field_type.__name__]}{"[]" * dimensions}'

            schema.append(field_definition)

        schema.append(f'PRIMARY KEY({", ".join(self.primary_key)})')

        return ', '.join(schema)

    @property
    def remote_fields(self) -> {str: type}:
        with self.connection:
            with self.connection.cursor() as cursor:
                if database_has_table(cursor, self.table):
                    cursor.execute(f'SELECT column_name, udt_name FROM information_schema.columns WHERE table_name=%s;',
                                   [self.table])
                    fields = {field[0]: field[1] for field in cursor.fetchall()}

                    for field, field_type in fields.items():
                        dimensions = field_type.count('_')
                        field_type = field_type.strip('_')

                        field_type = field_type.lower()
                        for python_type, postgres_type in POSTGRES_TYPES.items():
                            if postgres_type.lower() == field_type:
                                field_type = eval(python_type)
                                break
                        else:
                            for python_type, postgres_type in POSTGRES_TYPES.items():
                                if python_type.lower() in field_type:
                                    field_type = eval(python_type)
                                    break
                            else:
                                field_type = str

                        for _ in range(dimensions):
                            field_type = [field_type]
                        fields[field] = field_type

                    return fields

    @property
    def geometry_fields(self) -> {str: type}:
        return {field: field_type for field, field_type in self.fields.items() if field_type in GEOMETRY_TYPES}

    def __contains__(self, key: Any) -> bool:
        if not isinstance(key, Sequence) or isinstance(key, str):
            key = [key]
        with self.connection:
            with self.connection.cursor() as cursor:
                return database_table_has_record(cursor, self.table, dict(zip(self.primary_key, key)), self.primary_key)

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({repr(self.hostname)}, {repr(self.database)}, {repr(self.table)}, ' \
               f'{repr(self.fields)}, {repr(self.primary_key)}, {repr(self.crs)}, {repr(self.username)}, ' \
               f'{repr(self.password)})'


class InheritedTableError(Exception):
    pass


def open_tcp_port() -> int:
    open_socket = socket()
    open_socket.bind(('', 0))
    return open_socket.getsockname()[1]


def parse_record_values(record: {str: Any}, field_types: {str: type}) -> {str: Any}:
    """
    Parse the values in the given record into their respective field types.

    Parameters
    ----------
    record
        dictionary mapping fields to values
    field_types
        dictionary mapping fields to types

    Returns
    -------
    {str: Any}
        record with values parsed into their respective types
    """

    for field, value in record.items():
        if field in field_types:
            field_type = field_types[field]
            value_type = type(value)

            if value_type is not field_type and value is not None:
                if field_type is bool:
                    value = bool(value) if value_type is not str else literal_eval(value.capitalize())
                elif field_type is int:
                    value = int(value)
                elif field_type is float:
                    value = float(value)
                elif field_type is str:
                    value = str(value)
                elif value_type is str:
                    if field_type is list:
                        value = literal_eval(value)
                    elif field_type in (date, datetime):
                        value = datetime.strptime(value, '%Y%m%d')
                        if field_type is date:
                            value = value.date()
                    elif field_type in GEOMETRY_TYPES:
                        try:
                            value = wkb.loads(value, hex=True)
                        except WKBReadingError:
                            try:
                                value = wkt.loads(value)
                            except WKTReadingError:
                                try:
                                    value = wkb.loads(value)
                                except TypeError:
                                    value = shapely_shape(literal_eval(value))
                record[field] = value
    return record


def flatten_geometry(geometry: BaseGeometry) -> BaseGeometry:
    geometry_type = type(geometry)

    # strip 3rd dimension
    if 'POLYGON Z' in geometry.wkt:
        polygons = [polygon for polygon in geometry] if geometry_type is MultiPolygon else [geometry]
        for polygon_index, polygon in enumerate(polygons):
            exterior_2d = LinearRing([vertex[:2] for vertex in polygon.exterior.coords])
            interiors_2d = [LinearRing([vertex[:2] for vertex in interior.coords]) for interior in polygon.interiors]
            polygons[polygon_index] = Polygon(exterior_2d, interiors_2d)
        geometry = MultiPolygon(polygons) if geometry_type is MultiPolygon else Polygon(polygons[0])

    if not geometry.is_valid:
        geometry = geometry.buffer(0)
    return geometry


def database_has_table(cursor: psycopg2._psycopg.cursor, table: str) -> bool:
    """
    Whether the given table exists within the given database.

    Parameters
    ----------
    cursor
        psycopg2 cursor
    table
        name of table

    Returns
    -------
    bool
        whether table exists
    """

    cursor.execute(f'SELECT EXISTS(SELECT 1 FROM information_schema.tables WHERE table_name=%s);', [table.lower()])
    return cursor.fetchone()[0]


def database_table_has_record(cursor: psycopg2._psycopg.cursor, table: str, record: {str: Any},
                              primary_key: str = None) -> bool:
    """
    Whether the given record exists within the given table.

    Parameters
    ----------
    cursor
        psycopg2 cursor
    table
        name of table
    record
        dictionary record
    primary_key
        name of primary key

    Returns
    -------
    bool
        whether record exists in table
    """

    if primary_key is None:
        # cursor.execute(f'SELECT 1 FROM information_schema.table_constraints WHERE table_name=\'{table}\' AND
        # constraint_type= \'PRIMARY KEY\';')
        # primary_key_index = cursor.fetchone()[0] - 1
        #
        # cursor.execute(f'SELECT * FROM information_schema.columns WHERE table_name=\'{table}\';')
        # primary_key = cursor.fetchall()[primary_key_index]
        primary_key = list(record)[0]

    if not isinstance(primary_key, Sequence) or isinstance(primary_key, str):
        primary_key = [primary_key]

    values = []
    for key in primary_key:
        value = record[key]
        if type(value) is date:
            value = f'{value:%Y%m%d}'
        elif type(value) is datetime:
            value = f'{value:%Y%m%d %H%M%S}'
        values.append(value)
    values = tuple(values)

    cursor.execute(f'SELECT EXISTS(SELECT 1 FROM {table} WHERE ({", ".join(primary_key)}) = %s);', [values])
    return cursor.fetchone()[0]


def database_table_is_inherited(cursor: psycopg2._psycopg.cursor, table: str) -> bool:
    """
    Whether the given table is inherited.

    Parameters
    ----------
    cursor
        psycopg2 cursor
    table
        name of table

    Returns
    -------
    bool
        whether table is inherited
    """

    cursor.execute(f'SELECT EXISTS(SELECT 1 FROM pg_catalog.pg_inherits WHERE inhrelid=%s::regclass);', [f'public.{table}'])
    return cursor.fetchone()[0]


def database_table_columns(cursor: psycopg2._psycopg.cursor, table: str) -> {str: str}:
    """
    Get column names and data types of the given table, within the given database.

    Parameters
    ----------
    cursor
        psycopg2 cursor
    table
        name of table

    Returns
    -------
    {str: str}
        mapping of column names to the PostGres data type
    """

    cursor.execute(f'SELECT column_name, udt_name FROM information_schema.columns WHERE table_name=%s;', [table])
    return {record[0]: record[1] for record in cursor.fetchall()}


def postgis_geometry(geometry: Union[Point, LineString, Polygon, MultiPoint, MultiLineString, MultiPolygon],
                     epsg: int = None) -> str:
    """
    Convert Shapely geometry to a PostGIS geometry string.

    Parameters
    ----------
    geometry
        Shapely geometry
    epsg
        EPSG code of CRS

    Returns
    -------
    str
        PostGIS input string
    """

    if epsg is None:
        epsg = 4326

    return f'ST_SetSRID(\'{geometry.wkb_hex}\'::geometry, {epsg})'


def split_URL_port(url: str) -> (str, Union[str, None]):
    """
    Split the given URL into host and port, assuming port is appended after a colon.

    Parameters
    ----------
    url
        URL string

    Returns
    ----------
    str, Union[str, None]
        URL and port (if found)
    """

    port = None

    if url.count(':') > 0:
        url = url.split(':')
        if 'http' in url:
            url = ':'.join(url[:2])
            if len(url) > 2:
                port = int(url[2])
        else:
            url, port = url
            port = int(port)

    return url, port



from typing import Union

import aprslib


def parse_raw_aprs(raw_aprs: Union[str, dict]) -> dict:
    """
    Parse APRS fields from raw packet string.

    APRS format reference: http://www.aprs.org/doc/APRS101.PDF

    :param raw_aprs: raw APRS string
    :return: dictionary of APRS fields
    """

    if type(raw_aprs) is not dict:
        try:
            parsed_packet = aprslib.parse(raw_aprs)
        except aprslib.ParseError as error:
            raise PartialPacketError(str(error))
    else:
        parsed_packet = {
            'from'        : raw_aprs['srccall'],
            'to'          : raw_aprs['dstcall'],
            'path'        : raw_aprs['path'].split(','),
            'timestamp'   : raw_aprs['time'],
            'symbol'      : raw_aprs['symbol'][1:],
            'symbol_table': raw_aprs['symbol'][0],
            'latitude'    : float(raw_aprs['lat']),
            'longitude'   : float(raw_aprs['lng']),
            'altitude'    : float(raw_aprs['altitude']),
            'comment'     : raw_aprs['comment']
        }

    # parsed_packet = {'raw':str(raw_aprs)}
    #
    # if ':' not in raw_aprs:
    #     raise PartialPacketError('location data missing')
    #
    # parsed_packet['from'], raw_aprs = raw_aprs.split('>', 1)
    # parsed_packet['to'], raw_aprs = raw_aprs.split(',', 1)
    # parsed_packet['path'], raw_aprs = raw_aprs.split(':', 1)
    # parsed_packet['path'] = parsed_packet['path'].split(',')
    # parsed_packet['via'] = parsed_packet['path'][-1]
    #
    # parsed_packet['messagecapable'] = raw_aprs[0] in ['!', '/']
    #
    # try:
    #     raw_aprs = raw_aprs[2:]
    #
    #     # TODO add parsing for uncompressed coordinates
    #     lonlat_string = raw_aprs[:8]
    #     parsed_packet['latitude'] = decompress_aprs_lat(lonlat_string[:4])
    #     parsed_packet['longitude'] = decompress_aprs_lon(lonlat_string[4:])
    #     parsed_packet['symbol'] = raw_aprs[8]
    #
    #     raw_aprs = raw_aprs[9:]
    #
    #     # TODO add parsing for more altitude formats
    #     # convert altitude from feet to meters
    #     parsed_packet['altitude'] = int(re.findall('/A=.{6}', raw_aprs)[0][3:]) * 0.3048
    #
    #     parsed_packet['comment'] = re.split('/A=.{6}', raw_aprs)[-1]
    # except IndexError:
    #     raise PartialPacketError('packet terminated unexpectedly')

    return parsed_packet


class PartialPacketError(Exception):
    pass


def decompress_aprs_lon(compressed_lon: str) -> float:
    """
    Decode longitude string from APRS compressed format (shifted ASCII in base 91) to a float.

    :param compressed_lon: compressed APRS longitude string
    :return: longitude
    """

    converted_floats = []
    current_power = len(compressed_lon) - 1

    for character in compressed_lon:
        converted_floats.append(float(ord(character) - 33) * (91 ** current_power))
        current_power -= 1

    return -180 + (sum(converted_floats) / 190463)


def decompress_aprs_lat(compressed_lat: str) -> float:
    """
    Decode latitude string from APRS compressed format (shifted ASCII in base 91) to a float.

    :param compressed_lat: compressed APRS latitude string
    :return: latitude
    """

    converted_floats = []
    current_power = len(compressed_lat) - 1

    for character in compressed_lat:
        converted_floats.append(float(ord(character) - 33) * (91 ** current_power))
        current_power -= 1

    return 90 - (sum(converted_floats) / 380926)



from abc import ABC, abstractmethod
from datetime import datetime
import logging
from os import PathLike
from pathlib import Path
from typing import Any

from dateutil.parser import parse
import requests
from serial import Serial
from serial.tools import list_ports
from shapely.geometry import Point

from client import CREDENTIALS_FILENAME
from .database import DatabaseTable, database_table_has_record
from .packets import APRSPacket, LocationPacket
from .utilities import get_logger, read_configuration

LOGGER = get_logger('packetraven.connection')


class PacketConnection(ABC):
    def __init__(self, location: str):
        """
        Create a new generic packet connection.

        :param location: location of packets
        """

        self.location = location

    @property
    @abstractmethod
    def packets(self) -> [LocationPacket]:
        """
        List the most recent packets available from this connection.

        :return: list of packets
        """

        raise NotImplementedError

    @abstractmethod
    def __enter__(self):
        raise NotImplementedError

    @abstractmethod
    def __exit__(self, exc_type, exc_val, exc_tb):
        raise NotImplementedError

    @abstractmethod
    def close(self):
        raise NotImplementedError


class APRSPacketConnection(PacketConnection):
    def __init__(self, location: str, callsigns: [str]):
        """
        Create a new generic APRS packet connection.

        :param location:
        :param callsigns: list of callsigns to return from source
        """

        super().__init__(location)
        self.callsigns = callsigns

    @property
    @abstractmethod
    def packets(self) -> [APRSPacket]:
        """
        List the most recent packets available from this connection.

        :return: list of APRS packets
        """

        raise NotImplementedError


class SerialTNC(APRSPacketConnection):
    def __init__(self, serial_port: str = None, callsigns: [str] = None):
        """
        Connect to TNC over given serial port.

        :param serial_port: port name
        :param callsigns: list of callsigns to return from source
        """

        if serial_port is None or serial_port == 'auto':
            try:
                serial_port = next_available_port()
            except ConnectionError:
                raise ConnectionError('could not find TNC connected to serial')
        else:
            serial_port = serial_port.strip('"')

        self.connection = Serial(serial_port, baudrate=9600, timeout=1)
        super().__init__(self.connection.port, callsigns)

    @property
    def packets(self) -> [APRSPacket]:
        packets = [APRSPacket.from_raw_aprs(line, source=self.location) for line in self.connection.readlines()]
        if self.callsigns is not None:
            return [packet for packet in packets if packet.callsign in self.callsigns]
        else:
            return packets

    def __enter__(self):
        return self.connection

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def close(self):
        self.connection.close()

    def __repr__(self):
        return f'{self.__class__.__name__}("{self.location}")'


class TextFileTNC(APRSPacketConnection):
    def __init__(self, filename: PathLike = None, callsigns: str = None):
        """
        read APRS packets from a given text file where each line consists of the time sent (`YYYY-MM-DDTHH:MM:SS`) followed by
        the raw APRS string

        :param filename: path to text file
        :param callsigns: list of callsigns to return from source
        """

        if not isinstance(filename, Path):
            if isinstance(filename, str):
                filename = filename.strip('"')
            filename = Path(filename)

        super().__init__(filename, callsigns)
        self.connection = open(filename)

    @property
    def packets(self) -> [APRSPacket]:
        packets = []
        for line in self.connection.readlines():
            if len(line) > 0:
                try:
                    packet_time, raw_aprs = line.split(' ', 1)
                    packet_time = parse(packet_time)
                except:
                    raw_aprs = line
                    packet_time = datetime.now()
                raw_aprs = raw_aprs.strip()
                packets.append(APRSPacket.from_raw_aprs(raw_aprs, packet_time, source=self.location))
        if self.callsigns is not None:
            return [packet for packet in packets if packet.callsign in self.callsigns]
        else:
            return packets

    def __enter__(self):
        return self.connection

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def close(self):
        self.connection.close()

    def __repr__(self):
        return f'{self.__class__.__name__}("{self.location}")'


class APRSfiConnection(APRSPacketConnection):
    def __init__(self, callsigns: [str], api_key: str = None):
        """
        connect to https://aprs.fi

        :param callsigns: list of callsigns to return from source
        :param api_key: API key for aprs.fi
        """

        url = 'https://api.aprs.fi/api/get'
        if callsigns is None:
            raise ConnectionError(f'queries to {url} require a list of callsigns')
        super().__init__(url, callsigns)

        if api_key is None or api_key == '':
            configuration = read_configuration(CREDENTIALS_FILENAME)

            if 'APRS_FI' in configuration:
                api_key = configuration['APRS_FI']['api_key']
            else:
                raise ConnectionError(f'no APRS.fi API key specified')

        self.api_key = api_key

        if not self.connected:
            raise ConnectionError(f'no network connection')

    @property
    def api_key(self) -> str:
        return self.__api_key

    @api_key.setter
    def api_key(self, api_key: str):
        response = requests.get(f'{self.location}?name=OH2TI&what=wx&apikey={api_key}&format=json').json()
        if response['result'] == 'fail':
            raise ConnectionError(response['description'])
        self.__api_key = api_key

    @property
    def packets(self) -> [APRSPacket]:
        query = {
            'name'  : ','.join(self.callsigns),
            'what'  : 'loc',
            'apikey': self.api_key,
            'format': 'json'
        }

        query = '&'.join(f'{key}={value}' for key, value in query.items())

        response = requests.get(f'{self.location}?{query}').json()
        if response['result'] != 'fail':
            packets = [APRSPacket.from_raw_aprs(packet_candidate, source=self.location)
                       for packet_candidate in response['entries']]
        else:
            logging.warning(f'query failure "{response["code"]}: {response["description"]}"')
            packets = []

        return packets

    @property
    def connected(self) -> bool:
        """
        test network connection

        :return: whether current session has a network connection to https://api.aprs.fi/api/get
        """

        try:
            requests.get(self.location, timeout=2)
            return True
        except ConnectionError:
            return False

    def __enter__(self):
        if not self.connected:
            raise ConnectionError(f'No network connection.')

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def close(self):
        pass

    def __repr__(self):
        callsigns_string = ','.join(f'"{callsign}"' for callsign in self.callsigns)
        return f'{self.__class__.__name__}([{callsigns_string}], r"{self.api_key}")'


class PacketDatabaseTable(DatabaseTable, PacketConnection):
    __default_fields = {
        'time' : datetime,
        'x'    : float,
        'y'    : float,
        'z'    : float,
        'point': Point
    }

    def __init__(self, hostname: str, database: str, table: str, **kwargs):
        if 'fields' not in kwargs:
            kwargs['fields'] = {}
        if 'primary_key' not in kwargs:
            kwargs['primary_key'] = 'time'
        kwargs['fields'] = {**self.__default_fields, **kwargs['fields']}
        DatabaseTable.__init__(self, hostname, database, table, **kwargs)
        PacketConnection.__init__(self, f'postgresql://{self.hostname}:{self.port}/{self.database}/{self.table}')

    @property
    def packets(self) -> [LocationPacket]:
        return [LocationPacket(**{key: value for key, value in record.items() if key != 'point'}, source=self.location)
                for record in self.records]

    def __getitem__(self, key: Any) -> LocationPacket:
        record = super().__getitem__(key)
        record = {key.replace('packet_', ''): value for key, value in record.items()}
        return LocationPacket(**record, crs=self.crs)

    def __setitem__(self, time: datetime, packet: LocationPacket):
        if packet.crs != self.crs:
            packet.transform_to(self.crs)
        super().__setitem__(time, self.__packet_record(packet))

    def insert(self, packets: [APRSPacket]):
        for packet in packets:
            if packet.crs != self.crs:
                packet.transform_to(self.crs)
        records = [self.__packet_record(packet) for packet in packets]
        super().insert(records)

    def __contains__(self, packet: LocationPacket) -> bool:
        with self.connection:
            with self.connection.cursor() as cursor:
                return database_table_has_record(cursor, self.table, {key: packet[key.replace('packet_', '')]
                                                                      for key in self.primary_key},
                                                 self.primary_key)

    def __enter__(self):
        return self.connection

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def close(self):
        self.connection.close()
        if self.tunnel is not None:
            self.tunnel.stop()

    @staticmethod
    def __packet_record(packet: LocationPacket) -> {str: Any}:
        return {
            'time' : packet.time,
            'x'    : packet.coordinates[0],
            'y'    : packet.coordinates[1],
            'z'    : packet.coordinates[2],
            'point': Point(*packet.coordinates),
            **{f'packet_{field}': value for field, value in packet.attributes.items()}
        }


class APRSPacketDatabaseTable(PacketDatabaseTable, APRSPacketConnection):
    __aprs_fields = {
        'from'        : str,
        'to'          : str,
        'path'        : [str],
        'timestamp'   : str,
        'symbol'      : str,
        'symbol_table': str,
        'latitude'    : float,
        'longitude'   : float,
        'altitude'    : float,
        'comment'     : str
    }

    def __init__(self, hostname: str, database: str, table: str, callsigns: [str] = None, **kwargs):
        if 'fields' not in kwargs:
            kwargs['fields'] = self.__aprs_fields
        if 'primary_key' not in kwargs:
            kwargs['primary_key'] = ('time', 'packet_from')
        elif 'callsign' in kwargs['primary_key']:
            kwargs['primary_key'][kwargs['primary_key'].index('callsign')] = 'packet_from'
        kwargs['fields'] = {f'packet_{field}': field_type for field, field_type in kwargs['fields'].items()}
        PacketDatabaseTable.__init__(self, hostname, database, table, **kwargs)
        APRSPacketConnection.__init__(self, f'postgresql://{self.hostname}:{self.port}/{self.database}/{self.table}',
                                      callsigns)

    @property
    def packets(self) -> [APRSPacket]:
        return [
            APRSPacket(
                **{field if field in ['time', 'x', 'y', 'z'] else field.replace('packet_', ''): value
                   for field, value in record.items() if field not in ['point']},
                source=self.location
            ) for record in self.records
        ]

    def __getitem__(self, key: (datetime, str)) -> APRSPacket:
        packet = super().__getitem__(key)
        return APRSPacket(packet.time, *packet.coordinates, packet.crs, **packet.attributes)

    def __setitem__(self, key: (datetime, str), packet: APRSPacket):
        PacketDatabaseTable.__setitem__(self, key, packet)

    def __contains__(self, packet: APRSPacket) -> bool:
        with self.connection:
            with self.connection.cursor() as cursor:
                return database_table_has_record(cursor, self.table, {key: packet[key.replace('packet_', '')]
                                                                      for key in self.primary_key},
                                                 self.primary_key)

    def insert(self, packets: [APRSPacket]):
        PacketDatabaseTable.insert(self, packets)

    @property
    def records(self) -> [{str: Any}]:
        if self.callsigns is not None:
            return self.records_where({'callsign': self.callsigns})
        else:
            return self.records_where(None)

    @staticmethod
    def __packet_record(packet: LocationPacket) -> {str: Any}:
        return {
            'time'    : packet.time,
            'callsign': packet.callsign,
            'x'       : packet.coordinates[0],
            'y'       : packet.coordinates[1],
            'z'       : packet.coordinates[2],
            'point'   : Point(*packet.coordinates),
            **{f'packet_{field}': value for field, value in packet.attributes.items()}
        }


def available_ports() -> str:
    """
    Iterate over available serial ports.

    :return: port name
    """

    for com_port in list_ports.comports():
        yield com_port.device
    else:
        return None


def next_available_port() -> str:
    """
    Get next port in ports list.

    :return: port name
    """

    try:
        return next(available_ports())
    except StopIteration:
        raise ConnectionError('no open serial ports')
